From 38f39a9cb35cb5abbc478eadffbf98d2b6af34b5 Mon Sep 17 00:00:00 2001
From: Jarmo Torvinen <jarmo.torvinen@jutel.fi>
Date: Mon, 27 Apr 2009 13:57:29 +0300
Subject: [PATCH] Non-blocking file writer implementation

---
 modules/access_output/file.c |  196 ++++++++++++++++++++++++++++++++++++++----
 1 files changed, 178 insertions(+), 18 deletions(-)

diff --git a/modules/access_output/file.c b/modules/access_output/file.c
index 7639a0b..b5e6815 100644
--- a/modules/access_output/file.c
+++ b/modules/access_output/file.c
@@ -6,6 +6,7 @@
  *
  * Authors: Laurent Aimar <fenrir@via.ecp.fr>
  *          Eric Petit <titer@videolan.org>
+ *          Jarmo Torvinen <jarmo.torvinen@jutel.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -64,6 +65,8 @@ static void Close( vlc_object_t * );
 #define APPEND_TEXT N_("Append to file")
 #define APPEND_LONGTEXT N_( "Append to file if it exists instead " \
                             "of replacing it.")
+#define NB_BUFFER_TEXT N_("Non-blocking write buffer size")
+#define NB_BUFFER_LONGTEXT N_( "Maximum number of mux packets to buffer for non-blocking write. 0 = unlimited. Default = 5600.")
 
 vlc_module_begin ()
     set_description( N_("File stream output") )
@@ -75,6 +78,7 @@ vlc_module_begin ()
     add_shortcut( "stream" )
     add_bool( SOUT_CFG_PREFIX "append", 0, NULL, APPEND_TEXT,APPEND_LONGTEXT,
               true )
+    add_integer( SOUT_CFG_PREFIX "non-block-bufsize", 5600, NULL, NB_BUFFER_TEXT, NB_BUFFER_LONGTEXT, true );
     set_callbacks( Open, Close )
 vlc_module_end ()
 
@@ -83,17 +87,30 @@ vlc_module_end ()
  * Exported prototypes
  *****************************************************************************/
 static const char *const ppsz_sout_options[] = {
-    "append", NULL
+    "append", 
+    "non-block-bufsize",
+    NULL
 };
 
+static void* ThreadWrite( void * );
+static ssize_t WriteNonBlock( sout_access_out_t *, block_t * );
 static ssize_t Write( sout_access_out_t *, block_t * );
 static int Seek ( sout_access_out_t *, off_t  );
 static ssize_t Read ( sout_access_out_t *, block_t * );
 static int Control( sout_access_out_t *, int, va_list );
 
+
 struct sout_access_out_sys_t
 {
-    int i_handle;
+    VLC_COMMON_MEMBERS
+
+    block_fifo_t *p_fifo;
+    int         fd;
+
+    int64_t i_max_buf;
+    int i_discard_count;
+    vlc_thread_t  thread;
+
 };
 
 /*****************************************************************************
@@ -102,7 +119,7 @@ struct sout_access_out_sys_t
 static int Open( vlc_object_t *p_this )
 {
     sout_access_out_t   *p_access = (sout_access_out_t*)p_this;
-    int                 fd;
+    sout_access_out_sys_t *p_sys;
 
     config_ChainParse( p_access, SOUT_CFG_PREFIX, ppsz_sout_options, p_access->p_cfg );
 
@@ -114,17 +131,25 @@ static int Open( vlc_object_t *p_this )
 
     bool append = var_GetBool( p_access, SOUT_CFG_PREFIX "append" );
 
+    if( !( p_sys = calloc ( 1, sizeof( sout_access_out_sys_t ) ) ) )
+        return VLC_ENOMEM;
+    p_access->p_sys = p_sys;
+
+    p_sys->i_max_buf = (int64_t)1024 * var_GetInteger( p_access, SOUT_CFG_PREFIX "non-block-bufsize");
+    p_sys->i_discard_count = 0;
+    
+
     if( !strcmp( p_access->psz_path, "-" ) )
     {
 #ifndef UNDER_CE
 #ifdef WIN32
         setmode (fileno (stdout), O_BINARY);
 #endif
-        fd = dup (fileno (stdout));
+        p_sys->fd = dup (fileno (stdout));
         msg_Dbg( p_access, "using stdout" );
 #else
 #warning stdout is not supported on Windows Mobile, but may be used on Windows CE
-        fd = -1;
+        p_sys->fd = -1;
 #endif
     }
     else
@@ -132,40 +157,140 @@ static int Open( vlc_object_t *p_this )
         char *psz_tmp = str_format( p_access, p_access->psz_path );
         path_sanitize( psz_tmp );
 
-        fd = utf8_open( psz_tmp, O_RDWR | O_CREAT | O_LARGEFILE |
+        p_sys->fd = utf8_open( psz_tmp, O_RDWR | O_CREAT | O_LARGEFILE |
                         (append ? 0 : O_TRUNC), 0666 );
         free( psz_tmp );
     }
 
-    if (fd == -1)
+    if (p_sys->fd == -1)
     {
         msg_Err( p_access, "cannot open `%s' (%m)", p_access->psz_path );
+        free (p_sys);
+        return VLC_EGENERIC;
+    }
+
+    // check for named pipe access
+    struct stat stat_info;
+    bool nonblock;
+
+    if( fstat(p_sys->fd, &stat_info) == -1)
+    {
+        msg_Err( p_access, "cannot stat `%s' (%m)", p_access->psz_path );
+        free (p_sys);
         return VLC_EGENERIC;
     }
 
-    p_access->pf_write = Write;
+    nonblock = S_ISFIFO(stat_info.st_mode);
+
+    // for named pipes, enable non-blocking writing
+    p_access->pf_write = nonblock ? WriteNonBlock : Write;
     p_access->pf_read  = Read;
     p_access->pf_seek  = Seek;
     p_access->pf_control = Control;
-    p_access->p_sys    = (void *)(intptr_t)fd;
 
-    msg_Dbg( p_access, "file access output opened (%s)", p_access->psz_path );
+    msg_Dbg( p_access, "file access output opened (%s)%s", p_access->psz_path , nonblock ? " in nonblocking mode" : "");
     if (append)
-        lseek (fd, 0, SEEK_END);
+        lseek (p_sys->fd, 0, SEEK_END);
+
+
+    if( nonblock )
+    {
+        p_sys->p_fifo = block_FifoNew();
+
+        if( !p_sys->p_fifo )
+        {
+            msg_Err( p_access, "error while reserving fifo buffer" );
+            close (p_sys->fd);
+            free (p_sys);
+            return VLC_ENOMEM;
+        }
+
+        if( vlc_clone (&p_sys->thread, ThreadWrite, p_sys,
+                       VLC_THREAD_PRIORITY_OUTPUT))
+        {
+            msg_Err( p_access, "cannot spawn sout access thread" );
+            block_FifoRelease(p_sys->p_fifo);
+            close (p_sys->fd);
+            free (p_sys);
+            return VLC_EGENERIC;
+        }
+    }
 
     return VLC_SUCCESS;
 }
 
 /*****************************************************************************
+ * ThreadWrite: Write a packet to the file when it is possible
+ *****************************************************************************/
+static void* ThreadWrite( void *data )
+{
+    block_t *p_buffer;
+    sout_access_out_sys_t *p_sys = (sout_access_out_sys_t *) data;
+
+    for ( ;; )
+    {
+        // this will block until data available
+        p_buffer = block_FifoGet( p_sys->p_fifo );
+
+        if( p_buffer == NULL )
+            continue; /* forced wake-up */		
+
+        block_cleanup_push(p_buffer);
+        
+        for ( ;; )
+            {
+
+                ssize_t val = write (p_sys->fd, p_buffer->p_buffer, p_buffer->i_buffer);
+
+                if (val == -1)
+                {
+                    if (errno == EINTR)
+                        continue;
+
+                    block_ChainRelease (p_buffer);
+                    break;
+                }
+
+                if ((size_t)val < p_buffer->i_buffer)
+                {
+                    p_buffer->p_buffer += val;
+                    p_buffer->i_buffer -= val;
+                    continue;
+                }
+
+                // now we know it has been fully written, release the memory
+                block_ChainRelease (p_buffer); 
+                break;
+	        }
+        vlc_cleanup_pop();
+		
+		
+	}
+	return NULL;
+}
+
+
+/*****************************************************************************
  * Close: close the target
  *****************************************************************************/
 static void Close( vlc_object_t * p_this )
 {
     sout_access_out_t *p_access = (sout_access_out_t*)p_this;
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
 
-    close( (intptr_t)p_access->p_sys );
+    if( p_sys->p_fifo != NULL)
+    {
+        // we used non blocking mode, free all resources
+        block_FifoWake( p_sys->p_fifo );
+        vlc_cancel (p_sys->thread);
+        vlc_join (p_sys->thread, NULL);
+        block_FifoRelease(p_sys->p_fifo);
+    }
 
+    close( p_sys->fd );
     msg_Dbg( p_access, "file access output closed" );
+    free( p_sys );
+
 }
 
 static int Control( sout_access_out_t *p_access, int i_query, va_list args )
@@ -190,12 +315,15 @@ static int Control( sout_access_out_t *p_access, int i_query, va_list args )
  *****************************************************************************/
 static ssize_t Read( sout_access_out_t *p_access, block_t *p_buffer )
 {
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
     ssize_t val;
 
-    do
-        val = read( (intptr_t)p_access->p_sys, p_buffer->p_buffer,
-                    p_buffer->i_buffer );
-    while (val == -1 && errno == EINTR);
+    do {
+      
+        val = read( p_sys->fd, p_buffer->p_buffer,
+              p_buffer->i_buffer );
+    } while (val == -1 && errno == EINTR);
+
     return val;
 }
 
@@ -204,11 +332,12 @@ static ssize_t Read( sout_access_out_t *p_access, block_t *p_buffer )
  *****************************************************************************/
 static ssize_t Write( sout_access_out_t *p_access, block_t *p_buffer )
 {
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
     size_t i_write = 0;
 
     while( p_buffer )
     {
-        ssize_t val = write ((intptr_t)p_access->p_sys,
+        ssize_t val = write (p_sys->fd,
                              p_buffer->p_buffer, p_buffer->i_buffer);
         if (val == -1)
         {
@@ -235,9 +364,40 @@ static ssize_t Write( sout_access_out_t *p_access, block_t *p_buffer )
 }
 
 /*****************************************************************************
+* WriteNonBlock: NonBlocking buffered write on a file descriptor.
+*****************************************************************************/
+static ssize_t WriteNonBlock( sout_access_out_t *p_access, block_t *p_buffer )
+{
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
+    size_t i_total_size = 0;
+
+    // new block is always put on the fifo
+    i_total_size = block_FifoPut(p_sys->p_fifo, p_buffer);
+
+    // destroy oldest items from the buffer until it meets the memory limits
+    if( p_sys->i_max_buf > 0)
+        while( block_FifoCount(p_sys->p_fifo) > p_sys->i_max_buf )
+	    {
+	        p_buffer = block_FifoGet(p_sys->p_fifo);
+            block_ChainRelease (p_buffer);	
+
+            if( (p_sys->i_discard_count % 10000) == 0 )
+            {
+                msg_Err( p_access, "fifo buffer full, total number of packets discarded = %d ", p_sys->i_discard_count );
+            }
+            p_sys->i_discard_count++;
+        }
+
+    return i_total_size;
+}
+
+
+/*****************************************************************************
  * Seek: seek to a specific location in a file
  *****************************************************************************/
 static int Seek( sout_access_out_t *p_access, off_t i_pos )
 {
-    return lseek( (intptr_t)p_access->p_sys, i_pos, SEEK_SET );
+    sout_access_out_sys_t *p_sys = p_access->p_sys;
+
+    return lseek( p_sys->fd, i_pos, SEEK_SET );
 }
-- 
1.6.0.4

